#include "ProcessorBase.h"

namespace mdd{
    ProcessorBase::ProcessorBase(const std::string& base_config)
    {
        json jparse = json::parse(base_config);
        if (jparse.contains("configure"))
        {
            _base_config = jparse;
        }
        else {
            _base_config["configure"] = jparse;
        }
    }

    bool ProcessorBase::configure(const json& config) {
        json jconfig = config;
        bool sucess = false;
        auto jit = jconfig.find("GUI");
        if (jit != jconfig.end())
        {
            _base_config["GUI"].merge_patch(jit.value());
        }
        jit = jconfig.find("params");
        if (jit != jconfig.end())
        {
            
            auto jit2 = jit.value().find("inputs");
            if (jit2 != jit.value().end())
            {
                for (size_t i = 0; i < jit.value().size(); i++)
                {
                    inputs.push_back(std::make_shared<Parameter>());
                    if (auto input_ptr = inputs.back())
                    {
                        input_ptr->setParent(shared_from_this());
                        input_ptr->setName(jit2.value()["type"].get<std::string>());
                        input_ptr->setAppendix(jit2.value()["appendix"].get<int>());
                    }
                }
                sucess = true;
            }

            jit2 = jit.value().find("outputs");
            if (jit2 != jit.value().end())
            {
                for (size_t i = 0; i < jit.value().size(); i++)
                {
                    outputs.push_back(std::make_shared<Parameter>());
                    if (auto output_ptr = outputs.back())
                    {
                        output_ptr->setParent(shared_from_this());
                        output_ptr->setName(jit2.value()["type"].get<std::string>());
                        output_ptr->setAppendix(jit2.value()["appendix"].get<int>());
                    }
                }
                sucess = true;
            }
        }

        jit = jconfig.find("modules");
        if (jit != jconfig.end()) {
            auto regi = Registration();
            for (size_t i = 0; i < jit.value().size(); i++)
            {
                const json& jmodule = jit.value()[i];
                addModule(regi.generateModule(jmodule["id"].get<std::string>()));
                if (auto module_ptr = modules.back())
                {
                    module_ptr->setParent(shared_from_this());
                    module_ptr->setAppendix(jmodule["appendix"].get<int>());
                    module_ptr->load(jmodule["load"]);
                }
            }
            sucess = true;
        }

        jit = jconfig.find("connections");
        if (jit != jconfig.end()) {
            for (size_t i = 0; i < jit.value().size(); i++)
            {
                const json& jconn = jit.value()[i];
                connect(jconn["output"].get<std::string>(), jconn["inputs"].get<std::vector<std::string>>());
            }
            sucess = true;
        }

        jit = jconfig.find("configuration");
        if (jit != jconfig.end()) {
            sucess = configureChild(jit.value());
        }
        return sucess;
    }

    bool ProcessorBase::configureChild(const json& config) {
        return true;
    }

    const json& ProcessorBase::getConfiguration() {
        //update module apendix ? 
        return _base_config;
    }

    std::string ProcessorBase::getType() {
        return  type;
    }

    void ProcessorBase::setParent(std::shared_ptr<IModule> parent) {
        _parent = parent;
    }

    std::vector<std::string> ProcessorBase::getParentID() {
        std::vector<std::string> ret;
        if (auto parent_ptr = _parent.lock())
        {
            ret = parent_ptr->getParentID();
            ret.push_back(parent_ptr->getID());
        }
        return ret;
    }

    std::string ProcessorBase::getGeneratorKey() {
        return  key;
    }

    std::string ProcessorBase::setAppendix(int appendix) {
        _appendix = appendix;
        return  getID();
    }

    int ProcessorBase::getAppendix() {
        return _appendix;
    }

    std::string ProcessorBase::getID() {
        return key + std::to_string(_appendix);
    }


    std::string ProcessorBase::addModule(std::shared_ptr<IModule> module){
        modules.emplace_back(module);
        if (auto module_ptr = modules.back())
        {
            module_ptr->setParent(shared_from_this());
            module_ptr->setAppendix(module_counter++);
            return module_ptr->getID();
        }
        return "";
    }

    void ProcessorBase::removeModule(std::shared_ptr<IModule> module)
    {
        for (auto it = modules.begin(); it != modules.end(); ++it) {
            if ((*it) == module)
            {
                (*it)->disconnect();
                modules.erase(it);
                return;
            }
        }
    }

    std::vector<std::shared_ptr<Parameter>>& ProcessorBase::getInputParams() {
        return inputs;
    }

    std::vector<std::shared_ptr<Parameter>>& ProcessorBase::getOutputParams() {
        return outputs;
    }

    size_t ProcessorBase::getNumInputs() {
        return processor_inputs.size() + inputs.size();
    }

    size_t ProcessorBase::getNumOutputs() {
        return processor_outputs.size() + outputs.size();
    }
    std::vector< std::shared_ptr<IModule >>& ProcessorBase::getModules(){
        return modules;
    }

    size_t ProcessorBase::getNumModuls() {
        return modules.size();
    }

    std::shared_ptr<IModule> ProcessorBase::getModule(size_t index) {
        if (index < modules.size())
        {
            return modules[index];
        }
        return nullptr;
    }

    bool ProcessorBase::connect(const std::string& output_id, const std::vector<std::string>& input_ids)
    {
        std::size_t found = output_id.find_first_of("/");
        std::string str_module_id = output_id.substr(0, found);
        std::string str_output_id = output_id.substr( found + 1);
        std::shared_ptr<IOutput> out_ptr;
        if (str_module_id.empty())
        {
            out_ptr = getOutput(str_output_id);
        }
        else {
            for (size_t i = 0; i < modules.size(); i++)
            {
                if (auto module_ptr = modules[i])
                {
                    if (module_ptr->getID() == str_module_id)
                    {
                        out_ptr = module_ptr->getOutput(str_output_id);
                        break;
                    }
                }
            }
        }

        for (size_t i = 0; i < input_ids.size(); i++)
        {
            found = input_ids[i].find_first_of("/");
            str_module_id = input_ids[i].substr(0, found);
            std::string str_input_id = input_ids[i].substr(found + 1);

            if (str_module_id.empty())
            {
                auto in_ptr = getInput(str_input_id);
                in_ptr->connect(out_ptr);
            }
            else {
                for (size_t i = 0; i < modules.size(); i++)
                {
                    if (auto module_ptr = modules[i])
                    {
                        if (module_ptr->getID() == str_module_id)
                        {
                            auto in_ptr = module_ptr->getInput(str_input_id);
                            in_ptr->connect(out_ptr);
                            break;
                        }
                    }
                }
            }
        }
        return true;
    }

    void ProcessorBase::disconnect() {
        for (size_t i = 0; i < processor_inputs.size(); i++)
        {
            processor_inputs[i]->disconnect();
        }
        for (size_t i = 0; i < processor_outputs.size(); i++)
        {
            processor_outputs[i]->disconnect();
        }
        for (size_t i = 0; i < inputs.size(); i++)
        {
            inputs[i]->disconnect();
        }
        for (size_t i = 0; i < outputs.size(); i++)
        {
            outputs[i]->disconnect();
        }
    }

    void ProcessorBase::load(const json& j) {
        if (j.contains("ID"))
        {
            setAppendix(j["ID"]["appendix"].get<int>());
            j["ID"]["type"];

        }
        if (j.contains("key"))
        {

        }
        if (j.contains("type"))
        {

        }
        if (j.contains("configure"))
        {

        }

        if (j.contains("modules"))
        {
            Registration regi = Registration();
            for (json jmodule : j["modules"])
            {
                if (jmodule["type"].get<std::string>() == "module")
                {
                    addModule(regi.generateModule(jmodule["key"]));
                }
                else if (jmodule["type"].get<std::string>() == "processor")
                {
                    addModule(regi.generateProcessor(jmodule["key"]));
                }
                if (auto module_ptr = modules.back())
                {
                    module_ptr->setParent(shared_from_this());
                    module_ptr->load(jmodule);
                }
            }
        }

        if (j.contains("inputs"))
        {
            const size_t start_counter = processor_inputs.size();
            for (size_t i = start_counter; i < j["inputs"].size(); i++)
            {

            }
        }

        if (j.contains("outputs"))
        {
            const size_t start_counter = processor_outputs.size();
            for (size_t i = start_counter; i < j["outputs"].size(); i++)
            {

            }
        }

        if (j.contains("connections"))
        {
            for(json jcon : j["connections"])
            {
                json jout = jcon["output"];
                IModule::Ptr module_out_ptr = getModule(jout);
                IOutput::Ptr out_ptr = nullptr;
                std::string id = jout["name"].get<std::string>() + std::to_string(jout["appendix"].get<int>());
                for (size_t i = 0; i < module_out_ptr->getNumOutputs(); i++)
                {
                    auto out = module_out_ptr->getOutput(i);
                    if (out->getID() == id)
                    {
                        out_ptr = out;
                        break;
                    }
                }
                for (json jin : jcon["inputs"])
                {
                    IModule::Ptr module_in_ptr = getModule(jin);
                    IInput::Ptr in_ptr = nullptr;
                    std::string id = jin["name"].get<std::string>() + std::to_string(jin["appendix"].get<int>());
                    for (size_t i = 0; i < module_in_ptr->getNumInputs(); i++)
                    {
                        auto in = module_in_ptr->getInput(i);
                        if (in->getID() == id)
                        {
                            in_ptr = in;
                            break;
                        }
                    }
                    if (in_ptr != nullptr && out_ptr != nullptr)
                    {
                        in_ptr->connect(out_ptr);
                    }
                }
            }
        }

        //update appendix
        for (int i = 0; i < modules.size(); ++i)
        {
            modules[i]->setAppendix(i);
        }
    }

    json ProcessorBase::dump() {
        json ret = getConfiguration();
        ret["ID"] = getIdentifier();
        ret["type"] = type;
        ret["key"] = key;

        ret["params"];
        ret["params"]["inputs"];
        for (size_t i = 0; i < inputs.size(); i++)
        {
            if (auto in_ptr = inputs[i])
            {
                json sub;
                sub["type"] = in_ptr->getType();
                sub["appendix"] = in_ptr->getAppendix();
                ret["params"]["inputs"].push_back(sub);
            }
        }

        ret["params"]["outputs"];
        for (size_t i = 0; i < outputs.size(); i++)
        {
            if (auto out_ptr = outputs[i])
            {
                json sub;
                sub["type"] = out_ptr->getType();
                sub["appendix"] = out_ptr->getAppendix();
                ret["params"]["outputs"].push_back(sub);
            }
        }

        for (auto& in : processor_inputs)
        {
            in->setParent(shared_from_this());
            ret["inputs"].push_back(in->dump());
        }
        for (auto& in : inputs)
        {
            ret["inputs"].push_back(in->getInput(0)->dump());
        }

        for (auto& out : processor_outputs)
        {
            out->setParent(shared_from_this());
            ret["outputs"].push_back(out->dump());
        }
        for (auto& out : outputs)
        {
            ret["outputs"].push_back(out->getInput(0)->dump());
        }
        for (auto& mod : modules)
        {
            if (auto module_ptr = mod)
            {
                if (module_ptr->getType() == "module")
                {
                    ret["modules"].push_back(module_ptr->dump());
                }
                else
                {
                    ret["modules"].push_back(module_ptr->dump());
                }
            }
        }

        for (auto& mod : modules)
        {
            if (auto module_ptr = mod)
            {
                for (size_t j = 0; j < module_ptr->getNumOutputs(); j++)
                {
                    auto input_connections = module_ptr->getOutput(j)->getConnections();
                    if (!input_connections.empty())
                    {
                        json connect;
                        connect["output"] = module_ptr->getOutput(j)->getIdentifier();
                        for (size_t k = 0; k < input_connections.size(); k++)
                        {
                            connect["inputs"].push_back(input_connections[k]->getIdentifier());
                        }
                        ret["connections"].push_back(connect);
                    }
                }
            }
        }
        return ret;
    }

    json ProcessorBase::getIdentifier() {
        json jID;
        jID["key"] = key;
        jID["appendix"] = getAppendix();
        jID["type"] = type;
        jID["prefix"] = getParentID();
        return jID;
    }

    std::shared_ptr<IOutput> ProcessorBase::getOutput(size_t index) {
        if (index < processor_outputs.size())
        {
            return processor_outputs[index];
        }
        index -= processor_outputs.size();
        if (index < outputs.size())
        {
            return outputs[index]->getOutput(0);
        }
        return nullptr;
    }

    std::vector<std::shared_ptr<IInput>> ProcessorBase::getOptimizableInputs() {
        std::vector<std::shared_ptr<IInput>> ret;
        for (size_t i = 0; i < processor_inputs.size(); i++)
        {
            if (processor_inputs[i]->isOptimizable())
            {
                ret.push_back(processor_inputs[i]);
            }
        }
        for (size_t i = 0; i < inputs.size(); i++)
        {
            auto list = inputs[i]->getOptimizableInputs();
            ret.insert(ret.end(), list.begin(), list.end());
        }
        for (size_t i = 0; i < modules.size(); i++)
        {
            auto list = modules[i]->getOptimizableInputs();
            ret.insert(ret.end(), list.begin(), list.end());
        }
        return ret;
    }

    std::vector<std::shared_ptr<IOutput>> ProcessorBase::getOptimizableOutputs() {
        std::vector<std::shared_ptr<IOutput>> ret;
        for (size_t i = 0; i < processor_outputs.size(); i++)
        {
            if (processor_outputs[i]->isOptimizable())
            {
                ret.push_back(processor_outputs[i]);
            }
        }
        for (size_t i = 0; i < outputs.size(); i++)
        {
            auto list = outputs[i]->getOptimizableOutputs();
            ret.insert(ret.end(), list.begin(), list.end());
        }
        for (size_t i = 0; i < modules.size(); i++)
        {
            auto list = modules[i]->getOptimizableOutputs();
            ret.insert(ret.end(), list.begin(), list.end());
        }
        return ret;
    }
   
    std::shared_ptr<IOutput> ProcessorBase::getOutput(const json& jid){
        auto jid_pref = jid.find("prefix");
        if (jid_pref != jid.end())
        {
            auto mod = getModule(jid);
            if (mod == shared_from_this())
            {
                std::string sid = jid["key"].get<std::string>() + std::to_string(jid["appendix"].get<int>());
                for (auto& output : processor_outputs) {
                    if (output->getID() == sid) {
                        return output;
                    }
                }
                for (auto& output : outputs) {
                    if (output->getOutput(0)->getID() == sid) {
                        return output->getOutput(0);
                    }
                }
            }
            else
            {
                return mod->getOutput(jid);
            }
        }
        return nullptr;
    }

    std::shared_ptr<IInput> ProcessorBase::getInput(size_t index) {
        if (index < processor_inputs.size())
        {
            return processor_inputs[index];
        }
        index -= processor_inputs.size();
        if (index < inputs.size())
        {
            return inputs[index]->getInput(0);
        }
        return nullptr;
    }

    std::shared_ptr<IInput> ProcessorBase::getInput(const json& jid){
        auto jid_pref = jid.find("prefix");
        if (jid_pref != jid.end())
        {
            auto mod = getModule(jid);
            if (mod == shared_from_this())
            {
                std::string sid = jid["key"].get<std::string>() + std::to_string(jid["appendix"].get<int>());
                for (auto& input : processor_inputs) {
                    if (input->getID() == sid) {
                        return input;
                    }
                }
                for (auto& input : inputs) {
                    if (input->getInput(0)->getID() == sid) {
                        return input->getInput(0);
                    }
                }
            }
            else
            {
                return mod->getInput(jid);
            }
        }
        return nullptr;
    }

    std::shared_ptr<IModule> ProcessorBase::getModule(const json& jid){
        json jid_copy = jid;
        auto jid_typ = jid.find("type");
        auto jid_copy_pref = jid_copy.find("prefix");
        //check if called non recursive
        json own_prefix = getIdentifier()["prefix"];
        if (jid_copy_pref->dump() == own_prefix.dump())
        {
            jid_copy_pref->clear();
        }
        //recursive call
        if (!jid_copy_pref->empty())
        {
            jid_copy_pref->erase(jid_copy_pref->begin());
        }
        int final_size = 0;
        if (jid_typ != jid.end())
        {
            std::string sid;
            if (jid_copy_pref->empty())
            {
                if (*jid_typ == "input" || *jid_typ == "output")
                {
                    final_size = 1;
                    sid = jid_copy_pref->at(0).get<std::string>();
                }
                else
                {
                    sid = jid_copy["key"].get<std::string>() + std::to_string(jid_copy["appendix"].get<int>());
                }
            }
            else {
                sid = jid_copy_pref->begin()->get<std::string>();
            }

            IModule::Ptr module_ptr;
            for (auto& module : modules) {
                if (module->getID() == sid) {
                    module_ptr = module;
                    break;
                }
            }
            if (jid_copy_pref->size() > final_size)
            {
                if (ProcessorBase* proc_ptr = dynamic_cast<ProcessorBase*>(&(*module_ptr)))
                {
                    return proc_ptr->getModule(jid_copy);
                }
            }
            else if(module_ptr != nullptr)
            {
                return module_ptr;
            }
        }
        
        return shared_from_this();
        //return nullptr;
    }


}
//*/